from __future__ import annotations

import pathlib
from dataclasses import dataclass, field
from itertools import groupby
from typing import TYPE_CHECKING, Iterator

import numpy as np

from ada.core.utils import Counter
from ada.fem.formats.sesam.common import sesam_eltype_2_general
from ada.fem.formats.sesam.read import cards

if TYPE_CHECKING:
    from ada import Material, Section
    from ada.fem.formats.sesam.read.cards import DataCard
    from ada.fem.results.common import ElementFieldData, FEAResult, Mesh, NodalFieldData

FEM_SEC_NAME = Counter(prefix="FS")

STRESS_MAP = {
    1: ("SIGXX", "Normal Stress x-direction"),
    2: ("SIGYY", "Normal Stress y-direction"),
    4: ("TAUXY", "Shear stress in y-direction, yz-plane"),
}
FORCE_MAP = {
    1: ("NXX", "Normal force in x-direction, yz-plane"),
    2: ("NXY", "Shear force in y-direction, yz-plane"),
    3: ("NXZ", "Shear force in z-direction, yz-plane"),
    10: ("MXX", "Torsion moment around x-axis, yz-plane"),
    11: ("MXY", "Bending moment around y-axis, yz-plane"),
    12: ("MXZ", "Bending moment around z-axis, yz-plane"),
}
# Integration Point location.
# If Integration point is in nodal position
# (Int ID, Node ID, Thickness offset)
# Else
# (Int ID, (X relative Coord, Y relative Coord, Thickness offset))

INT_LOCATIONS = {
    24: [
        (0, 0, -0.5),
        (1, 1, -0.5),
        (2, (0.5, 0.5, -0.5)),
        (3, 2, -0.5),
        (4, 3, -0.5),
        (5, 0, 0.5),
        (6, 1, 0.5),
        (7, (0.5, 0.5, 0.5)),
        (8, 2, 0.5),
        (9, 3, 0.5),
    ],
    25: [
        (0, 0, -0.5),
        (1, 1, -0.5),
        (2, (0.5, 0.5), -0.5),
        (3, 2, -0.5),
        (4, 0, 0.5),
        (5, 1, 0.5),
        (6, (0.5, 0.5), 0.5),
        (7, 2, 0.5),
    ],
    15: [(0, 0), (1, 0.5), (2, 1)],
}
OTHER_CARDS = [
    cards.GUNIVEC,
    cards.TDSECT,
    cards.TDMATER,
    cards.MISOSEL,
    cards.MORSMEL,
    cards.TDSETNAM,
    cards.GSETMEMB,
    cards.TDRESREF,
]
SECTION_CARDS = [cards.GIORH, cards.GBOX]
RESULT_CARDS = [
    cards.RVNODDIS,
    cards.RVSTRESS,
    cards.RDPOINTS,
    cards.RDSTRESS,
    cards.RDIELCOR,
    cards.RDRESREF,
    cards.RVFORCES,
    cards.RDFORCES,
]


@dataclass
class SifReader:
    file: Iterator

    nodes: np.ndarray = None
    node_ids: np.ndarray = None
    elements: list[tuple] = None
    results: list[tuple] = field(default_factory=list)

    skipped_flags: list[str] = field(default_factory=list)

    _last_line: str | None = None

    _gelref1: list = None
    _other: dict[str, list] = field(default_factory=dict)
    _sections: dict[str, list] = field(default_factory=dict)

    def _read_single_line_statements(self, startswith: str, first_line: str):
        first_vals = [float(x) for x in first_line.split()[1:]]
        yield first_vals

        while True:
            line = next(self.file)
            stripped = line.strip()

            if stripped.startswith(startswith) is False:
                self._last_line = line
                break

            yield [float(x) for x in stripped.split()[1:]]

    def _read_multi_line_statements(self, startswith: str, first_line: str):
        curr_elements = [float(x) for x in first_line.split()[1:]]
        while True:
            stripped = next(self.file).strip()

            if stripped.startswith(startswith) is False and stripped[0].isnumeric() is False and stripped[0] != "-":
                self._last_line = stripped
                yield curr_elements
                break

            if stripped.startswith(startswith):
                yield curr_elements
                curr_elements = [float(x) for x in stripped.split()[1:]]
            else:
                curr_elements += [float(x) for x in stripped.split()]

    def read_gcoords(self, first_line: str):
        for data in self._read_single_line_statements("GCOORD", first_line):
            yield data

    def read_gnodes(self, first_line: str):
        for data in self._read_single_line_statements("GNODE", first_line):
            yield data[:2]

    def read_gelmnts(self, first_line: str):
        elno_id, eltyp, members = cards.GELMNT1.get_indices_from_names(["elno", "eltyp", "nids"])
        for data in self._read_multi_line_statements("GELMNT1", first_line):
            yield data[eltyp], data[elno_id], data[members:]

    def read_results(self, result_variable: str, first_line: str) -> tuple:
        for data in self._read_multi_line_statements(result_variable, first_line):
            yield data

    def get_sections(self) -> dict[int, Section]:
        from ada import Section

        sec_map: dict[str, cards.DataCard] = {s.name: s for s in SECTION_CARDS}
        td_sect_map = self.get_tdsect_map()

        sections = dict()
        for sec_name, sec_data in self._sections.items():
            sec_card = sec_map[sec_name]
            sm = cards.SEC_MAP[sec_name]
            sec_type = sm[0]
            keys = [x[0] for x in sm[1]]
            for s in sec_data:
                res = sec_card.get_data_map_from_names(["geono", *keys], s)
                sec_id = int(float(res["geono"]))
                prop_map = {ada_n: res[ses_n] for ses_n, ada_n in sm[1]}
                sec_tdsect = td_sect_map.get(sec_id)
                if sec_tdsect is None:
                    raise ValueError(f"TDSECT is not set for section ID {sec_id}")
                sec_name = sec_tdsect[-1]
                sec = Section(name=sec_name, sec_id=sec_id, sec_type=sec_type, **prop_map)
                sections[sec_id] = sec

        return sections

    def get_sets(self):
        from ada.fem import FemSet

        member_map = self.get_gsetmemb()
        if member_map is None:
            return None
        set_map = self.get_tdsetnam_map()
        istype_i, isorig_i = cards.GSETMEMB.get_indices_from_names(["ISTYPE", "ISORIG"])
        sets = dict()
        for set_id, props in member_map.items():
            eltype = props[istype_i]
            set_type = "nset" if eltype == 1 else "elset"
            set_name = set_map[set_id][-1]
            members = props[isorig_i:]
            sets[set_name] = FemSet(set_name, members, set_type=set_type)
        return sets

    def get_materials(self) -> dict[int, Material]:
        from ada import Material
        from ada.materials.metals import CarbonSteel

        mat_map = {int(x[1]): x[-1] for x in self._other.get("TDMATER", [])}
        isotrop_mats = {int(x[0]): x for x in self._other.get("MISOSEL", [])}
        anisotrop_mats = {int(x[0]): x for x in self._other.get("MORSMEL", [])}

        sm_anisotropic = cards.MAT_MAP["MORSMEL"]
        anisotropic_keys = [s[0] for s in sm_anisotropic]

        sm_isotropic = cards.MAT_MAP["MISOSEL"]
        isotropic_keys = [s[0] for s in sm_isotropic]

        materials = dict()
        for mat_id, mat_data in isotrop_mats.items():
            res = cards.MISOSEL.get_data_map_from_names(isotropic_keys, mat_data)
            mat_name = mat_map[mat_id]
            prop_map = {ada_n: res[ses_n] for ses_n, ada_n in sm_isotropic}
            grade = get_grade(prop_map["sig_y"])
            model = CarbonSteel(grade=grade, **prop_map)
            mat = Material(name=mat_name, mat_id=mat_id, mat_model=model)
            materials[mat_id] = mat

        for mat_id, mat_data in anisotrop_mats.items():
            res = cards.MORSMEL.get_data_map_from_names(anisotropic_keys, mat_data)
            mat_name = mat_map[mat_id]
            prop_map = {ada_n: res[ses_n] for ses_n, ada_n in sm_anisotropic}
            grade = get_grade(prop_map["sig_y"])
            model = CarbonSteel(grade=grade, **prop_map)
            mat = Material(name=mat_name, mat_id=mat_id, mat_model=model)
            materials[mat_id] = mat

        return materials

    def get_vectors(self) -> dict[int, list]:
        res = self._other.get("GUNIVEC")
        if res is None:
            return None
        return {x[0]: x[1:] for x in res}

    def get_gelref(self):
        return self._gelref1

    def eval_flags(self, line: str):
        stripped = line.strip()
        is_skipped_flag = False

        # Nodes
        if stripped.startswith(cards.GCOORD.name):
            self.nodes = np.array(list(self.read_gcoords(stripped)))
        elif stripped.startswith(cards.GNODE.name):
            self.node_ids = np.array(list(self.read_gnodes(stripped)))

        # Elements
        elif stripped.startswith(cards.GELMNT1.name):
            self.elements = list(self.read_gelmnts(stripped))

        elif stripped[0].isnumeric() is False and stripped[0] != "-":
            is_skipped_flag = True

        # Sections
        if stripped.startswith(cards.GELREF1.name):
            self._gelref1 = list(self.iter_card(cards.GELREF1, self.file, stripped))

        for other_card in OTHER_CARDS:
            if stripped.startswith(other_card.name):
                if other_card.name in self._other.keys():
                    self._other[other_card.name] += list(self.iter_card(other_card, self.file, stripped))
                else:
                    self._other[other_card.name] = list(self.iter_card(other_card, self.file, stripped))

        for sec_card in SECTION_CARDS:
            if stripped.startswith(sec_card.name):
                self._sections[sec_card.name] = list(self.iter_card(sec_card, self.file, stripped))

        # if len(self._other) > 0 and self._gelref1 is not None and len(self._sections) > 0:
        #     self.read_fem_sections()

        # Results
        for res_card in RESULT_CARDS:
            if stripped.startswith(res_card.name):
                self.results.append((res_card.name, list(self.read_results(res_card.name, stripped))))
                is_skipped_flag = False

        if is_skipped_flag:
            flag = stripped.split()[0]
            if flag not in self.skipped_flags:
                self.skipped_flags.append(flag)

    def iter_card(self, datacard: DataCard, f: Iterator, curr_line: str):
        curr_elements = [float(x) for x in curr_line.split()[1:]]
        startswith = datacard.name
        n_field = None
        if datacard.components[0] == "nfield":
            n_field = int(curr_elements[0])

        while True:
            num_el = len(curr_elements)
            stripped = next(f).strip()

            if n_field is not None and num_el >= n_field:
                result = datacard.str_to_proper_types(stripped)
                if len(result) == 0:
                    curr_elements += [stripped]
                else:
                    if datacard.name in ("TDSECT", "TDSETNAM", "TDMATER", "TDRESREF"):
                        curr_elements += result
                yield curr_elements
                break

            if n_field is None:
                if stripped.startswith(startswith) is False and datacard.is_numeric(stripped) is False:
                    yield curr_elements
                    break

                if stripped.startswith(startswith):
                    yield curr_elements
                    curr_elements = [float(x) if datacard.is_numeric(x) else x for x in stripped.split()[1:]]
                    continue

            curr_elements += datacard.str_to_proper_types(stripped)

        self._last_line = stripped

    def load(self):
        while True:
            try:
                if self._last_line is not None:
                    line = self._last_line
                    self._last_line = None
                else:
                    line = next(self.file)
                self.eval_flags(line)
            except StopIteration:
                break

    def get_result(self, name: str) -> list:
        result = list(filter(lambda x: x[0] == name, self.results))

        if len(result) > 1:
            raise NotImplementedError("")

        return result

    def get_rdpoints_map(self) -> dict:
        rdpoints = self.get_result(cards.RDPOINTS.name)[0][1]
        return {x[2]: x for x in rdpoints[1:]}

    def get_rdstress_map(self) -> dict:
        rdstress = self.get_result(cards.RDSTRESS.name)
        return {int(x[1]): tuple([int(i) for i in x[3:]]) for x in rdstress[0][1]}

    def get_rdielcor_map(self) -> dict:
        rdielcor = self.get_result(cards.RDIELCOR.name)
        return {int(x[1]): x[2:] for x in rdielcor[0][1]}

    def get_rdforces_map(self) -> dict:
        rdforces = self.get_result(cards.RDFORCES.name)
        return {int(x[1]): tuple([int(i) for i in x[3:]]) for x in rdforces[0][1]}

    def get_tdsect_map(self):
        res = self._other.get(cards.TDSECT.name)
        if res is None:
            return None
        return {x[1]: x for x in res}

    def get_tdsetnam_map(self):
        res = self._other.get(cards.TDSETNAM.name)
        if res is None:
            return None
        return {int(x[1]): x for x in res}

    def get_gsetmemb(self):
        res = self._other.get(cards.GSETMEMB.name)
        if res is None:
            return None
        return {int(x[1]): [int(i) for i in x] for x in res}

    def get_rdresref(self):
        res = self.get_result(cards.RDRESREF.name)[0][1]
        if res is None:
            return None
        return {int(x[1]): [int(i) for i in x] for x in res}

    def get_tdresref(self):
        res = self._other.get(cards.TDRESREF.name)
        if res is None:
            return None
        return {int(x[1]): x for x in res}


def read_sif_file(sif_file: str | pathlib.Path) -> FEAResult:
    sif_file = pathlib.Path(sif_file)

    with open(sif_file, "r") as f:
        sif = SifReader(f)
        sif.load()

    s2m = Sif2Mesh(sif)
    fea_results = s2m.convert(sif_file)

    return fea_results


@dataclass
class Sif2Mesh:
    sif: SifReader

    mesh: Mesh = None

    def convert(self, sif_file) -> FEAResult:
        from ada.fem.results.common import FEAResult, FEATypes

        self.mesh = self.get_sif_mesh()
        results = self.get_sif_results()
        rnames = self.get_result_name_map()

        return FEAResult(
            sif_file.stem,
            FEATypes.SESAM,
            results=results,
            mesh=self.mesh,
            results_file_path=sif_file,
            step_name_map=rnames,
        )

    def get_sif_mesh(self) -> Mesh:
        from ada.fem.results.common import (
            ElementBlock,
            ElementInfo,
            FEATypes,
            FemNodes,
            Mesh,
        )
        from ada.fem.shapes.definitions import ShapeResolver

        sif = self.sif

        nodes = FemNodes(coords=sif.nodes[:, 1:], identifiers=sif.node_ids[:, 0])
        elem_blocks = []
        for eltype, elements in groupby(sif.elements, key=lambda x: x[0]):
            elem_type = int(eltype)
            elem_data = list(elements)
            general_elem_type = sesam_eltype_2_general(elem_type)
            num_nodes = ShapeResolver.get_el_nodes_from_type(general_elem_type)
            elem_identifiers = np.array([x[1] for x in elem_data], dtype=int)
            elem_node_refs = np.array([x[2][:num_nodes] for x in elem_data], dtype=float)
            res = sesam_eltype_2_general(elem_type)
            elem_info = ElementInfo(type=res, source_software=FEATypes.SESAM, source_type=elem_type)
            elem_blocks.append(
                ElementBlock(elem_info=elem_info, node_refs=elem_node_refs, identifiers=elem_identifiers)
            )

        sets = self.sif.get_sets()
        sections = self.sif.get_sections()
        materials = self.sif.get_materials()
        vectors = self.sif.get_vectors()
        elem_refs = cards.GELREF1.cast_to_np(["elno", "matno", "geono", "transno"], self.sif.get_gelref())

        return Mesh(
            elements=elem_blocks,
            nodes=nodes,
            sections=sections,
            materials=materials,
            vectors=vectors,
            elem_data=elem_refs,
            sets=sets,
        )

    def get_sif_results(self) -> list[ElementFieldData | NodalFieldData]:
        result_blocks = self.get_nodal_data()
        result_blocks += self.get_field_data()

        return result_blocks

    def get_result_name_map(self):
        tdresref = self.sif.get_tdresref()
        rdresref = self.sif.get_rdresref()
        return {key: tdresref[value[1]][-1] for key, value in rdresref.items()}

    def get_nodal_data(self) -> list[NodalFieldData]:
        return get_nodal_results(self.sif.get_result(cards.RVNODDIS.name)[0][1])

    def get_field_data(self) -> list[ElementFieldData | NodalFieldData]:
        sif = self.sif
        field_results = []

        if len(sif.get_result(cards.RVSTRESS.name)) > 0:
            field_results += self.get_field_shell_data()

        if len(sif.get_result(cards.RVFORCES.name)) > 0:
            field_results += self.get_field_line_data()

        return field_results

    def get_field_line_data(self):
        ires_i, ielno_i, irforc_i = cards.RVFORCES.get_indices_from_names(["ires", "ielno", "irforc|"])
        nsp_i, eltyp_i = cards.RDPOINTS.get_indices_from_names(["nsp", "ieltyp"])

        rdpoints_map = self.sif.get_rdpoints_map()

        def keyfunc(x):
            iielno = x[ielno_i]
            rdpoints_res = rdpoints_map[iielno]
            _nsp = int(rdpoints_res[nsp_i])
            _elem_type = int(rdpoints_res[eltyp_i])
            return x[ires_i], _nsp, _elem_type, x[irforc_i]

        field_results = []

        force_res_name = cards.RVFORCES.name
        for (ires, nsp, elem_type, irforc), rv_forces in groupby(
            sorted(self.sif.get_result(force_res_name)[0][1][1:], key=keyfunc), key=keyfunc
        ):
            if elem_type not in (15,):
                continue
            field_data = self._get_line_field_data(rv_forces, int(ires), int(irforc), elem_type, nsp)
            field_results.append(field_data)

        return field_results

    def _get_line_field_data(self, rv_forces, ires, irforc, elem_type, nsp) -> ElementFieldData:
        from ada.fem.results.common import ElementFieldData

        rdforces_map = self.sif.get_rdforces_map()
        force_types = [FORCE_MAP[c][0] for c in rdforces_map[irforc]]
        data = np.array(list(_iter_line_forces(rv_forces, rdforces_map, nsp)))
        return ElementFieldData(
            "FORCES",
            int(ires),
            components=force_types,
            values=data,
            field_pos=ElementFieldData.field_pos.INT,
            int_positions=INT_LOCATIONS[elem_type],
        )

    def get_field_shell_data(self):
        sif = self.sif
        ires_i, irstrs_i, iielno_i = cards.RVSTRESS.get_indices_from_names(["ires", "irstrs", "iielno"])
        nsp_i, eltyp_i = cards.RDPOINTS.get_indices_from_names(["nsp", "ieltyp"])

        rdpoints_map = sif.get_rdpoints_map()

        def keyfunc(x):
            iielno = x[iielno_i]
            rdpoints_res = rdpoints_map[iielno]
            _nsp = int(rdpoints_res[nsp_i])
            _elem_type = int(rdpoints_res[eltyp_i])
            return x[ires_i], _nsp, _elem_type, x[irstrs_i]

        field_results = []

        for (ires, nsp, elem_type, irstrs), rv_stresses in groupby(
            sorted(sif.get_result(cards.RVSTRESS.name)[0][1][1:], key=keyfunc), key=keyfunc
        ):
            if elem_type not in (25, 24):
                continue

            field_data = self._get_shell_field_data(rv_stresses, ires, irstrs, elem_type, nsp)
            field_results.append(field_data)

        return field_results

    def _get_shell_field_data(self, rv_stresses, ires, irstrs, elem_type: int, nsp) -> ElementFieldData:
        from ada.fem.results.common import ElementFieldData

        rdstress_map = self.sif.get_rdstress_map()
        stress_types = [STRESS_MAP[c][0] for c in rdstress_map[irstrs]]
        data = np.array(list(_iter_shell_stress(rv_stresses, rdstress_map, nsp)))
        return ElementFieldData(
            "STRESS",
            int(ires),
            components=stress_types,
            values=data,
            field_pos=ElementFieldData.field_pos.INT,
            int_positions=INT_LOCATIONS[elem_type],
        )

    def get_int_positions(self, rdpoints_res, nlay_i, nsptra_i) -> list:
        iielno_i, ieltyp_i, icoref_i, ijkdim_i = cards.RDPOINTS.get_indices_from_names(
            ["iielno", "ieltyp", "icoref", "ijkdim"]
        )
        iielno = int(rdpoints_res[iielno_i])
        elem = self.mesh.get_elem_by_id(iielno)
        points = [n.p for n in elem.nodes]
        nsptra_len = int(rdpoints_res[nsptra_i] * 9)
        # rmat = np.array(rdpoints_res[-nsptra_len:]).reshape((3, 3))
        nox_data = _get_rdpoints_nox_data(rdpoints_res, nlay_i, nsptra_len)
        p0 = points[0]
        relative_points = np.zeros((len(nox_data), 3))
        for i, v in enumerate(nox_data.values()):
            rel_p = p0 - np.array(v)
            relative_points[i] = rel_p
        # rdielcor_map = sif.get_rdielcor_map()
        # ijkdim = rdpoints_res[ijkdim_i]
        # nok = int(ijkdim / 10000)
        # noj = int((ijkdim % 10000) / 100)
        # noi = int(ijkdim % 100)

        # rdielcor_res = rdielcor_map[int(rdpoints_res[icoref_i])]
        # rdielcor_res_copy = copy.deepcopy(rdielcor_res)
        # gamma = [rdielcor_res_copy.pop() for x in range(0, nok)]
        # beta = [rdielcor_res_copy.pop() for x in range(0, noj)]
        # alpha = [rdielcor_res_copy.pop() for x in range(0, noi)]
        el_type = int(rdpoints_res[ieltyp_i])

        return INT_LOCATIONS[el_type]


def get_nodal_results(res) -> list[NodalFieldData]:
    from ada.fem.results.common import NodalFieldData

    comps = "U1|", "U2|", "U3|", "U4|", "U5|", "U6|"
    indices = cards.RVNODDIS.get_indices_from_names(["inod", *comps])
    nid = indices[0]
    start = indices[1]
    stop = indices[-1] + 1
    results = []
    for ires, data in groupby(res[1:], key=lambda x: x[1]):
        field_data = np.asarray([(x[nid], *x[start:stop]) for x in data], dtype=float)
        fd = NodalFieldData(cards.RVNODDIS.name, int(ires), [x.replace("|", "") for x in comps], field_data)
        results.append(fd)

    return results


def _get_rdpoints_nox_data(rdpoints_res, nlay_i, nsptra_len):
    nox_data = rdpoints_res[nlay_i + 1 : -nsptra_len]
    nox_data_clean = dict()
    data_iter = iter(nox_data)
    init_int_point = next(data_iter)
    if init_int_point == -1:
        remaining_el = 0
    else:
        remaining_el = 3

    curr_int_id = int(init_int_point)
    nox_data_clean[curr_int_id] = []

    while True:
        try:
            curr_int_point = next(data_iter)
        except StopIteration:
            break

        if remaining_el > 0:
            nox_data_clean[curr_int_id].append(curr_int_point)
            remaining_el -= 1
        else:
            curr_int_id = int(curr_int_point)
            if curr_int_id == -1:
                remaining_el = 0
            else:
                remaining_el = 3
                nox_data_clean[curr_int_id] = []

    return nox_data_clean


def _iter_shell_stress(rv_stresses: Iterator, rdstress_map, nsp) -> Iterator:
    ires_i, iielno_i, ispalt_i, irstrs_i = cards.RVSTRESS.get_indices_from_names(["ires", "iielno", "ispalt", "irstrs"])
    for rv_stress in rv_stresses:
        iielno = int(rv_stress[iielno_i])
        irstrs = int(rv_stress[irstrs_i])
        rdstress_res = rdstress_map[irstrs]
        data = np.array(rv_stress[irstrs_i + 1 :])

        reshaped_data = data.reshape((nsp, len(rdstress_res)))

        for i, data_per_int in enumerate(reshaped_data, start=1):
            yield iielno, i, *data_per_int


def _iter_line_forces(rv_forces: Iterator, rdforces_map, nsp) -> Iterator:
    ires_i, iielno_i, ispalt_i, irforc_i = cards.RVFORCES.get_indices_from_names(["ires", "ielno", "ispalt", "irforc|"])

    for rv_force in rv_forces:
        iielno = int(rv_force[iielno_i])
        irforc = int(rv_force[irforc_i])
        rdstress_res = rdforces_map[irforc]
        data = np.array(rv_force[irforc_i + 1 :])
        for i, data_per_int in enumerate(data.reshape((nsp, len(rdstress_res))), start=1):
            yield iielno, i, *data_per_int


def get_grade(sig_y, tol=1):
    from ada.materials.metals import CarbonSteel

    if abs(sig_y - 420e6) < tol:
        grade = CarbonSteel.TYPES.S420
    elif abs(sig_y - 355e6) < tol:
        grade = CarbonSteel.TYPES.S355
    else:
        grade = "NA"

    return grade



from __future__ import annotations

from typing import TYPE_CHECKING

import ifcopenshell.geom

from ada import Assembly, Shape
from ada.config import get_logger

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


def import_ifc_shape(product: ifcopenshell.entity_instance, name, ifc_store: IfcStore):
    logger.info(f'importing Shape "{name}"')
    color, opacity = get_colour(product, ifc_store.assembly)

    return Shape(
        name,
        None,
        guid=product.GlobalId,
        ifc_store=ifc_store,
        units=ifc_store.assembly.units,
        colour=color,
        opacity=opacity,
    )


def get_ifc_geometry(ifc_elem, settings):
    pdct_shape = ifcopenshell.geom.create_shape(settings, inst=ifc_elem)

    if pdct_shape is None:
        print(f'Unable to import geometry for ifc element "{ifc_elem}"')
        return pdct_shape, None, None

    geom = get_geom(ifc_elem, settings)
    r, g, b, alpha = pdct_shape.styles[0]  # the shape color

    colour = None if (r, g, b) == (-1, -1, -1) else (r, g, b)

    return geom, colour, alpha


def get_colour(product: ifcopenshell.entity_instance, assembly: Assembly) -> None | tuple:
    styles = []
    f = assembly.ifc_store.f
    for geo in f.traverse(product):
        if hasattr(geo, "StyledByItem") is False:
            continue
        if len(geo.StyledByItem) != 0:
            cstyle = geo.StyledByItem[0].Styles[0]
            if cstyle not in styles:
                styles.append(cstyle)

    if len(styles) == 0:
        logger.info(f'No style associated with IFC element "{product}"')
        return None, 1.0

    if len(styles) > 1:
        logger.warning(f"Multiple styles associated to element {product}. Choosing arbitrarily style @ index=0")

    style = styles[0]
    colour_rgb = list(filter(lambda x: x.is_a("IfcColourRgb"), f.traverse(style)))
    transparency = list(filter(lambda x: x.is_a("IfcSurfaceStyleRendering"), f.traverse(style)))

    if len(colour_rgb) == 0:
        logger.warning(f'ColourRGB not found for IFC product "{product}"')
        return None, 1.0

    opacity = 1.0 if len(transparency) == 0 else transparency[0].Transparency
    rgb = colour_rgb[0].Red, colour_rgb[0].Green, colour_rgb[0].Blue

    return rgb, opacity


def get_geom(ifc_elem, settings):
    from ifcopenshell.geom.occ_utils import shape_tuple
    from OCC.Core import BRepTools
    from OCC.Core.TopoDS import TopoDS_Compound

    try:
        pdct_shape = ifcopenshell.geom.create_shape(settings, inst=ifc_elem)
    except RuntimeError:
        print(f'unable to parse ifc_elem "{ifc_elem}"')
        return

    if isinstance(pdct_shape, shape_tuple):
        shape = pdct_shape[1]
    else:
        shape = pdct_shape.solid

    if type(shape) is not TopoDS_Compound:
        brep_data = pdct_shape.solid.brep_data
        ss = BRepTools.BRepTools_ShapeSet()
        ss.ReadFromString(brep_data)
        nb_shapes = ss.NbShapes()
        occ_shape = ss.Shape(nb_shapes)
    else:
        occ_shape = shape
    return occ_shape



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ada.fem import StepImplicit
from ada.fem.loads import Load

from ..write_loads import write_load

if TYPE_CHECKING:
    from ada.concepts.spatial import Part


@dataclass
class StatNonLin:
    name: str
    part: Part
    load: Load

    @property
    def sec_str(self):
        sec_str = ""
        if len(self.part.fem.sections.lines) > 0 or len(self.part.fem.sections.shells) > 0:
            sec_str = "\n    CARA_ELEM=element,"
        return sec_str

    def get_bc_str(self):
        from ada.fem.exceptions.model_definition import NoBoundaryConditionsApplied

        part = self.part
        all_boundary_conditions = part.fem.bcs
        assembly = part.get_assembly()
        if assembly != part:
            for bc in part.get_assembly().fem.bcs:
                if bc not in all_boundary_conditions:
                    all_boundary_conditions.append(bc)

        if len(all_boundary_conditions) == 0:
            raise NoBoundaryConditionsApplied("No boundary condition is found for the specified model")

        bc_str = ""
        for bc in all_boundary_conditions:
            bc_str += f"_F(CHARGE={bc.name}),"

        return bc_str

    def write(self):
        return f"""{self.name} = STAT_NON_LINE(
    MODELE=model,
    CHAM_MATER=material,{self.sec_str}
    COMPORTEMENT=(_F(DEFORMATION="PETIT", TOUT="OUI")),
    CONVERGENCE=_F(ARRET="OUI", ITER_GLOB_MAXI=8,),
    EXCIT=({self.get_bc_str()}_F(CHARGE={self.load.name}, FONC_MULT=bc_step)),
    INCREMENT=_F(LIST_INST=timeInst),
    ARCHIVAGE=_F(LIST_INST=timeReel),
)"""


@dataclass
class PostCalc:
    stat_non_line: StatNonLin
    part: Part

    def write(self):
        post_calc_str = f"""
{self.stat_non_line.name} = CALC_CHAMP(
    reuse={self.stat_non_line.name}, RESULTAT={self.stat_non_line.name},
    CONTRAINTE=("EFGE_ELNO", "EFGE_NOEU", "SIGM_ELNO"),
    DEFORMATION=("EPSI_ELNO", "EPSP_ELNO"),
)"""
        if len(self.part.fem.sections.solids) > 0:
            return post_calc_str

        return (
            post_calc_str
            + """

stress = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
        NOM_CHAM=('SIGM_ELNO', ),
        NUME_COUCHE=1
    ),
    RESULTAT=result
)

stress = CALC_CHAMP(
    reuse=stress,
    CONTRAINTE=('SIGM_NOEU', ),
    RESULTAT=stress
)

strain = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('EPSI_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)

strainP = POST_CHAMP(
    EXTR_COQUE=_F(
        NIVE_COUCHE='MOY',
    NOM_CHAM=('EPSP_ELNO', ),
    NUME_COUCHE=1),
    RESULTAT=result
)"""
        )


@dataclass
class ImprResu:
    stat_nl: StatNonLin
    part: Part

    @property
    def post_calc_include(self):
        if len(self.part.fem.sections.solids) > 0:
            return ""

        return """_F(
            NOM_CHAM=("SIGM_ELNO", "SIGM_NOEU"),
            NOM_CHAM_MED=("STRESSES_ELEM", "STRESSES_NODES"),
            RESULTAT=stress,
        ),
        _F(
            NOM_CHAM=("EPSI_ELNO",),
            NOM_CHAM_MED=("STRAINS_ELEM",),
            RESULTAT=strain,
        ),
        _F(
            NOM_CHAM=("EPSP_ELNO",),
            NOM_CHAM_MED=("PLASTIC_STRAINS_ELEM",),
            RESULTAT=strainP,
        ),"""

    def write(self):
        return f"""IMPR_RESU(
    RESU=(
        _F(
            NOM_CHAM=("DEPL", "EFGE_ELNO", "EFGE_NOEU"),
            NOM_CHAM_MED=("DISP", "GEN_FORCES_ELEM", "GEN_FORCES_NODES"),
            RESULTAT={self.stat_nl.name},
        ),
        {self.post_calc_include}
    ),
    UNITE=80,
)"""


def step_static_nonlin_str(step: StepImplicit, part: Part) -> str:
    from ada.fem.exceptions.model_definition import NoLoadsApplied

    load_str = "\n".join(list(map(write_load, step.loads)))
    if len(step.loads) == 0:
        raise NoLoadsApplied(f"No loads are applied in step '{step}'")

    load = step.loads[0]

    stat_non_line = StatNonLin("result", part, load)
    stat_non_line_str = stat_non_line.write()
    post_calc = PostCalc(stat_non_line, part)
    post_calc_str = post_calc.write()
    iresu = ImprResu(stat_non_line, part)
    iresu_str = iresu.write()

    return f"""
{load_str}

timeReel = DEFI_LIST_REEL(DEBUT=0.0, INTERVALLE=_F(JUSQU_A=1.0, NOMBRE=10))
timeInst = DEFI_LIST_INST(METHODE="AUTO", DEFI_LIST=_F(LIST_INST=timeReel))
bc_step = DEFI_FONCTION(NOM_PARA="INST", VALE=(0.0, 0.0, 1.0, 1.0))

{stat_non_line_str}
{post_calc_str}
{iresu_str}
"""



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import h5py
import meshio
import numpy as np

from ada.config import get_logger
from ada.fem import StepEigen
from ada.fem.elements import ElemShape
from ada.fem.formats.code_aster.read.reader import med_to_fem

if TYPE_CHECKING:
    from ada.fem.results import Results
    from ada.fem.results.eigenvalue import EigenDataSummary

logger = get_logger()


def get_eigen_data(rmed_file) -> EigenDataSummary:
    from ada.fem.results.eigenvalue import EigenDataSummary, EigenMode

    with h5py.File(rmed_file) as f:
        modes = f.get("CHA/modes___DEPL")
        eigen_modes = []

        for mname, m in modes.items():
            mode = m.attrs["NDT"]
            freq = m.attrs["PDT"]
            eigen_modes.append(EigenMode(mode, f_hz=freq))

    return EigenDataSummary(eigen_modes)


def get_eigen_frequency_deformed_meshes(rmed_file):
    fem = med_to_fem(rmed_file, "temp")

    with h5py.File(rmed_file) as f:
        modes = f.get("CHA/modes___DEPL")
        nodes = fem.nodes.to_np_array()
        eig_deformed_meshes = []

        for mname, m in modes.items():
            res = m["NOE"]["MED_NO_PROFILE_INTERNAL"]["CO"][()]
            dofs = res.reshape(len(fem.nodes), 6)
            eig_deformed_meshes.append(nodes + np.delete(dofs, np.s_[2:5], 1))
            mode = m.attrs["NDT"]
            freq = m.attrs["PDT"]
            print(mode, freq)

    # TODO: Figure out what kind of information is needed for animating frames in threejs/blender
    return fem, eig_deformed_meshes


def read_code_aster_results(results: "Results", file_ref: pathlib.Path, overwrite):
    if results.assembly is not None and isinstance(results.assembly.fem.steps[0], StepEigen):
        results.eigen_mode_data = get_eigen_data(file_ref)

    fem = med_to_fem(file_ref, "temp")
    if any([x.type == ElemShape.TYPES.shell.TRI7 for x in fem.elements.shell]):
        logger.error("Meshio does not support 7 node Triangle elements yet")
        return None

    if any([x.type == ElemShape.TYPES.shell.QUAD9 for x in fem.elements.shell]):
        logger.error("Meshio does not support 9 node QUAD elements yet")
        return None

    return meshio.read(file_ref, "med")



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Iterable, List

import numpy as np

from ada import FEM
from ada.fem.utils import is_line_elem

from .renderer_occ import occ_shape_to_faces

if TYPE_CHECKING:
    from ada.visualize.concept import ObjectMesh


def from_cache(hdf_file, guid):
    import h5py

    from .concept import ObjectMesh

    with h5py.File(hdf_file, "r") as f:
        res = f["VISMESH"].get(guid, None)
        if res is None:
            return None
        colour = list(res.attrs["COLOR"])
        translation = res.attrs["TRANSLATION"]
        return ObjectMesh(
            guid,
            res["INDEX"][()],
            res["POSITION"][()],
            res["NORMAL"][()],
            colour,
            translation=translation,
        )


def convert_obj_to_poly(obj, quality=1.0, render_edges=False, parallel=False):
    geom = obj.solid
    np_vertices, poly_indices, np_normals, _ = occ_shape_to_faces(geom, quality, render_edges, parallel)
    obj_buffer_arrays = np.concatenate([np_vertices, np_normals], 1)
    buffer, indices = np.unique(obj_buffer_arrays, axis=0, return_index=False, return_inverse=True)
    return dict(
        guid=obj.guid,
        index=indices.astype(int).tolist(),
        position=buffer.flatten().astype(float).tolist(),
        color=[*obj.colour_norm, obj.opacity],
        instances=[],
    )


def get_vertices_from_fem(fem: FEM) -> np.ndarray:
    return np.asarray([n.p for n in fem.nodes.nodes], dtype="float32")


def get_faces_from_fem(fem: FEM):
    ids = []
    for el in fem.elements.elements:
        if is_line_elem(el):
            continue
        for f in el.shape.faces:
            # Convert to indices, not id
            ids += [[int(e.id - 1) for e in f]]
    return ids


def get_edges_from_fem(fem: FEM):
    ids = []
    for el in fem.elements.elements:
        for f in el.shape.edges_seq:
            # Convert to indices, not id
            ids += [[int(el.nodes[e].id - 1) for e in f]]
    return ids


def organize_by_colour(objects: Iterable[ObjectMesh]) -> Dict[tuple, List[ObjectMesh]]:
    colour_map: Dict[tuple, List[ObjectMesh]] = dict()
    for obj in objects:
        colour = tuple(obj.color) if obj.color is not None else None
        if colour not in colour_map.keys():
            colour_map[colour] = []
        colour_map[colour].append(obj)
    return colour_map


def merge_mesh_objects(list_of_objects: Iterable[ObjectMesh]) -> ObjectMesh:
    from ada.ifc.utils import create_guid

    from .concept import ObjectMesh

    obj_mesh = ObjectMesh(
        create_guid(),
        np.array([], dtype=int),
        np.array([], dtype=float),
        np.array([], dtype=float),
    )

    for obj in list_of_objects:
        obj_mesh += obj

    return obj_mesh



from typing import Union


def get_instance_name(obj, written_on_assembly_level: bool) -> str:
    from ada import FEM, Assembly, Node, Part

    parent: Union[FEM, Part] = obj.parent
    p = parent.parent if type(parent) is FEM else parent
    obj_ref = obj.id if type(obj) is Node else obj.name

    if type(p) is Assembly:
        obj_on_assembly_level = True
    else:
        obj_on_assembly_level = False

    if written_on_assembly_level is True and obj_on_assembly_level is False:
        if obj.parent is None:
            raise AttributeError
        return f"{obj.parent.instance_name}.{obj_ref}"
    else:
        return str(obj_ref)



from __future__ import annotations

import numpy as np

from ada import (
    Penetration,
    PrimBox,
    PrimCyl,
    PrimExtrude,
    PrimRevolve,
    PrimSphere,
    PrimSweep,
    Shape,
)
from ada.base.units import Units
from ada.core.constants import O, X, Z
from ada.core.vector_utils import unit_vector, vector_length
from ada.ifc.utils import (
    add_colour,
    create_ifc_placement,
    create_ifcextrudedareasolid,
    create_IfcFixedReferenceSweptAreaSolid,
    create_ifcindexpolyline,
    create_ifcpolyline,
    create_ifcrevolveareasolid,
    create_local_placement,
    tesselate_shape,
    to_real,
)

from .write_curves import write_curve_poly


def write_ifc_shape(shape: Shape):
    if shape.parent is None:
        raise ValueError("Parent cannot be None for IFC export")

    a = shape.parent.get_assembly()
    f = a.ifc_store.f

    owner_history = a.ifc_store.owner_history
    parent = f.by_guid(shape.parent.guid)
    schema = f.wrapped_data.schema

    shape_placement = create_local_placement(f, relative_to=parent.ObjectPlacement)

    if isinstance(shape, (PrimBox, PrimCyl, PrimExtrude, PrimRevolve, PrimSphere, PrimSweep)):
        ifc_shape = generate_parametric_solid(shape, f)
    else:
        tol = Units.get_general_point_tol(a.units)
        serialized_geom = tesselate_shape(shape.geom(), schema, tol)
        ifc_shape = f.add(serialized_geom)

    # Add colour
    if shape.colour is not None:
        add_colour(f, ifc_shape.Representations[0].Items[0], str(shape.colour), shape.colour)

    ifc_elem = f.create_entity(
        str(shape.ifc_class.value),
        GlobalId=shape.guid,
        OwnerHistory=owner_history,
        Name=shape.name,
        ObjectType=None,
        ObjectPlacement=shape_placement,
        Representation=ifc_shape,
    )

    return ifc_elem


def generate_parametric_solid(shape: Shape | PrimSphere, f):
    a = shape.parent.get_assembly()
    body_context = a.ifc_store.get_context("Body")

    param_solid_map = {
        PrimSphere: generate_ifc_PrimSphere_geom,
        PrimBox: generate_ifc_box_geom,
        PrimCyl: generate_ifc_cylinder_geom,
        PrimExtrude: generate_ifc_prim_extrude_geom,
        PrimRevolve: generate_ifc_prim_revolve_geom,
        PrimSweep: generate_ifc_prim_sweep_geom,
    }

    ifc_geom_converter = param_solid_map.get(type(shape), None)
    if ifc_geom_converter is None:
        raise NotImplementedError(f'Shape type "{type(shape)}" is not yet supported for export to IFC')

    solid_geom = ifc_geom_converter(shape, f)

    if type(shape) is Penetration:
        raise ValueError(f'Penetration type "{shape}" is not yet supported')

    shape_representation = f.create_entity("IfcShapeRepresentation", body_context, "Body", "SweptSolid", [solid_geom])
    ifc_shape = f.create_entity("IfcProductDefinitionShape", None, None, [shape_representation])

    return ifc_shape


def generate_ifc_PrimSphere_geom(shape: PrimSphere, f):
    """Create IfcSphere from primitive PrimSphere"""
    opening_axis_placement = create_ifc_placement(f, to_real(shape.cog), Z, X)
    return f.createIfcSphere(opening_axis_placement, float(shape.radius))


def generate_ifc_box_geom(shape: PrimBox, f):
    """Create IfcExtrudedAreaSolid from primitive PrimBox"""
    p1 = shape.p1
    p2 = shape.p2
    points = [
        p1,
        (p1[0], p2[1], p1[2]),
        (p2[0], p2[1], p1[2]),
        (p2[0], p1[1], p1[2]),
    ]
    depth = p2[2] - p1[2]
    polyline = create_ifcpolyline(f, points)
    profile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)
    opening_axis_placement = create_ifc_placement(f, O, Z, X)
    return create_ifcextrudedareasolid(f, profile, opening_axis_placement, (0.0, 0.0, 1.0), depth)


def generate_ifc_cylinder_geom(shape: PrimCyl, f):
    """Create IfcExtrudedAreaSolid from primitive PrimCyl"""
    p1 = shape.p1
    p2 = shape.p2
    r = shape.r

    vec = np.array(p2) - np.array(p1)
    uvec = unit_vector(vec)
    vecdir = to_real(uvec)

    cr_dir = np.array([0, 0, 1])

    if vector_length(abs(uvec) - abs(cr_dir)) == 0.0:
        cr_dir = np.array([1, 0, 0])

    perp_dir = np.cross(uvec, cr_dir)

    if vector_length(perp_dir) == 0.0:
        raise ValueError("Perpendicular dir cannot be zero")

    create_ifc_placement(f, to_real(p1), vecdir, to_real(perp_dir))

    opening_axis_placement = create_ifc_placement(f, to_real(p1), vecdir, to_real(perp_dir))

    depth = vector_length(vec)
    profile = f.createIfcCircleProfileDef("AREA", shape.name, None, r)
    return create_ifcextrudedareasolid(f, profile, opening_axis_placement, Z, depth)


def generate_ifc_prim_extrude_geom(shape: PrimExtrude, f):
    """Create IfcExtrudedAreaSolid from primitive PrimExtrude"""
    # https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/link/annex-e.htm
    # polyline = self.create_ifcpolyline(self.file, [p[:3] for p in points])
    normal = shape.poly.normal
    h = shape.extrude_depth
    points = [tuple(x.astype(float).tolist()) for x in shape.poly.seg_global_points]
    seg_index = shape.poly.seg_index
    polyline = create_ifcindexpolyline(f, points, seg_index)
    profile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)
    opening_axis_placement = create_ifc_placement(f, O, Z, X)
    return create_ifcextrudedareasolid(f, profile, opening_axis_placement, [float(n) for n in normal], h)


def generate_ifc_prim_revolve_geom(shape: PrimRevolve, f):
    """Create IfcRevolveAreaSolid from primitive PrimRevolve"""
    # https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/link/annex-e.htm
    # 8.8.3.28 IfcRevolvedAreaSolid

    revolve_axis = [float(n) for n in shape.revolve_axis]
    revolve_origin = [float(x) for x in shape.revolve_origin]
    revolve_angle = shape.revolve_angle
    points = [tuple(x.astype(float).tolist()) for x in shape.poly.seg_global_points]
    seg_index = shape.poly.seg_index
    polyline = create_ifcindexpolyline(f, points, seg_index)
    profile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)
    opening_axis_placement = create_ifc_placement(f, O, Z, X)
    return create_ifcrevolveareasolid(
        f,
        profile,
        opening_axis_placement,
        revolve_origin,
        revolve_axis,
        revolve_angle,
    )


def generate_ifc_prim_sweep_geom(shape: PrimSweep, f):
    sweep_curve = write_curve_poly(shape.sweep_curve)
    outer_curve = write_curve_poly(shape.profile_curve_outer)
    profile = f.create_entity("IfcArbitraryClosedProfileDef", "AREA", None, outer_curve)
    ifc_xdir = f.create_entity("IfcDirection", [float(x) for x in shape.profile_curve_outer.xdir])
    opening_axis_placement = create_ifc_placement(f, O, Z, X)

    return create_IfcFixedReferenceSweptAreaSolid(f, sweep_curve, profile, opening_axis_placement, None, None, ifc_xdir)



from ada.concepts.curves import CurvePoly


def write_curve_poly(curve: CurvePoly):
    a = curve.parent.parent.get_assembly()
    f = a.ifc_store.f

    ifc_segments = []
    for seg_ind in curve.seg_index:
        if len(seg_ind) == 2:
            ifc_segments.append(f.createIfcLineIndex(seg_ind))
        elif len(seg_ind) == 3:
            ifc_segments.append(f.createIfcArcIndex(seg_ind))
        else:
            raise ValueError("Unrecognized number of values")

    # TODO: Investigate using 2DLists instead is it could reduce complexity?
    points = [tuple(x.astype(float).tolist()) for x in curve.seg_global_points]
    ifc_point_list = f.createIfcCartesianPointList3D(points)
    segindex = f.createIfcIndexedPolyCurve(ifc_point_list, ifc_segments)
    return segindex



import numpy as np

from ada import Plate
from ada.core.constants import O, X, Z
from ada.ifc.utils import (
    add_colour,
    create_ifc_placement,
    create_ifcindexpolyline,
    create_ifcpolyline,
    create_local_placement,
)


def write_ifc_plate(plate: Plate):
    if plate.parent is None:
        raise ValueError("Ifc element cannot be built without any parent element")

    a = plate.get_assembly()
    ifc_store = a.ifc_store
    f = ifc_store.f

    owner_history = ifc_store.owner_history
    parent = f.by_guid(plate.parent.guid)

    xvec = plate.poly.xdir
    zvec = plate.poly.normal
    yvec = np.cross(zvec, xvec)

    # Wall creation: Define the wall shape as a polyline axis and an extruded area solid
    plate_placement = create_local_placement(f, relative_to=parent.ObjectPlacement)

    tra_mat = np.array([xvec, yvec, zvec])
    t_vec = [0, 0, plate.t]
    origin = np.array(plate.poly.placement.origin)
    res = origin + np.dot(tra_mat, t_vec)
    polyline = create_ifcpolyline(f, [origin.astype(float).tolist(), res.tolist()])
    axis_representation = f.createIfcShapeRepresentation(ifc_store.get_context("Axis"), "Axis", "Curve2D", [polyline])
    extrusion_placement = create_ifc_placement(f, O, Z, X)
    points = [(float(n[0]), float(n[1]), float(n[2])) for n in plate.poly.seg_global_points]
    seg_index = plate.poly.seg_index
    polyline = create_ifcindexpolyline(f, points, seg_index)
    # polyline = plate.create_ifcpolyline(f, point_list)
    ifcclosedprofile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)

    ifcdir = f.createIfcDirection(zvec.astype(float).tolist())
    ifcextrudedareasolid = f.createIfcExtrudedAreaSolid(ifcclosedprofile, extrusion_placement, ifcdir, plate.t)

    body = f.createIfcShapeRepresentation(ifc_store.get_context("Body"), "Body", "SolidModel", [ifcextrudedareasolid])

    product_shape = f.createIfcProductDefinitionShape(None, None, [axis_representation, body])

    ifc_plate = f.createIfcPlate(
        plate.guid,
        owner_history,
        plate.name,
        plate.name,
        None,
        plate_placement,
        product_shape,
        None,
    )

    # Add colour
    if plate.colour is not None:
        add_colour(f, ifcextrudedareasolid, str(plate.colour), plate.colour)

    # Material
    ifc_store.writer.associate_elem_with_material(plate.material, ifc_plate)

    return ifc_plate



from ada import FEM
from ada.fem import Csys

from . import cards
from .helper_utils import get_set_from_assembly


def get_lcsys_from_bulk(bulk_str: str, parent: FEM) -> dict[str, Csys]:
    """
    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-orientation.htm#simakey-r-orientation

    """
    lcsysd = dict()
    for m in cards.orientation.regex.finditer(bulk_str):
        d = m.groupdict()
        name = d["name"].replace('"', "")
        defi = d.get("definition", "COORDINATES")
        system = d.get("system", "RECTANGULAR")
        if defi is None or defi.upper() == "COORDINATES":
            coords = [tuple(float(d[x]) for x in ["ax", "ay", "az"]), tuple(float(d[x]) for x in ["bx", "by", "bz"])]
            if d["cx"] is not None:
                coords += [(float(d["cx"]), float(d["cy"]), float(d["cz"]))]
            lcsysd[name] = Csys(name, system=system, coords=coords, parent=parent)
        elif defi.upper() == "NODES":
            nodes = []
            for n in ["ax", "ay", "az"]:
                nodes += [get_set_from_assembly(d[n], parent, "nset")]
            lcsysd[name] = Csys(name, system=system, definition=defi, nodes=nodes, parent=parent)
        else:
            raise NotImplementedError(f'Orientation definition "{defi}" is not yet supported')

    return lcsysd



from typing import TYPE_CHECKING

from ada.fem.interactions import ContactTypes
from ada.fem.steps import Step, StepExplicit

from .helper_utils import get_instance_name

if TYPE_CHECKING:
    from ada import FEM, Assembly
    from ada.fem import Interaction, InteractionProperty


def interact_str(fem: "FEM"):
    return "\n".join([interaction_str(interact) for interact in fem.interactions.values()])


def interaction_str(interaction: "Interaction") -> str:
    # Allowing Free text to be parsed directly through interaction class.
    if "aba_bulk" in interaction.metadata.keys():
        return interaction.metadata["aba_bulk"]

    contact_mod = interaction.metadata["contact_mod"] if "contact_mod" in interaction.metadata.keys() else "NEW"
    contact_incl = (
        interaction.metadata["contact_inclusions"]
        if "contact_inclusions" in interaction.metadata.keys()
        else "ALL EXTERIOR"
    )

    top_str = f"**\n** Interaction: {interaction.name}"
    if interaction.type == ContactTypes.SURFACE:
        adjust_par = interaction.metadata.get("adjust", None)
        geometric_correction = interaction.metadata.get("geometric_correction", None)
        small_sliding = interaction.metadata.get("small_sliding", None)

        first_line = "" if small_sliding is None else f", {small_sliding}"

        if issubclass(type(interaction.parent), Step):
            step = interaction.parent
            first_line += "" if type(step) is StepExplicit else f", type={interaction.surface_type}"
        else:
            first_line += f", type={interaction.surface_type}"

        if interaction.constraint is not None:
            first_line += f", mechanical constraint={interaction.constraint}"

        if adjust_par is not None:
            first_line += f", adjust={adjust_par}" if adjust_par is not None else ""

        if geometric_correction is not None:
            first_line += f", geometric correction={geometric_correction}"

        return f"""{top_str}
*Contact Pair, interaction={interaction.interaction_property.name}{first_line}
{get_instance_name(interaction.surf1, True)}, {get_instance_name(interaction.surf2, True)}"""
    else:
        return f"""{top_str}\n*Contact, op={contact_mod}
*Contact Inclusions, {contact_incl}
*Contact Property Assignment
 ,  , {interaction.interaction_property.name}"""


def interaction_prop_str(int_prop: "InteractionProperty") -> str:
    iprop_str = f"*Surface Interaction, name={int_prop.name}\n"

    # Friction
    iprop_str += f"*Friction\n{int_prop.friction},\n"

    # Behaviours
    tab_str = (
        "\n" + "\n".join(["{:>12.3E},{:>12.3E}".format(d[0], d[1]) for d in int_prop.tabular])
        if int_prop.tabular is not None
        else ""
    )
    iprop_str += f"*Surface Behavior, pressure-overclosure={int_prop.pressure_overclosure}{tab_str}"

    return iprop_str.rstrip()


def int_prop_str(fem: "FEM"):
    iprop_str = "\n".join([interaction_prop_str(iprop) for iprop in fem.intprops.values()])
    smoothings = fem.metadata.get("surf_smoothing", None)
    if smoothings is not None:
        iprop_str += "\n"
        for smooth in smoothings:
            name = smooth["name"]
            iprop_str += f"*Surface Smoothing, name={name}\n"
            iprop_str += smooth["bulk"] + "\n"
    return iprop_str


def eval_interactions(assembly: "Assembly", analysis_dir):
    if len(assembly.fem.steps) > 0:
        initial_step = assembly.fem.steps[0]
        if type(initial_step) is StepExplicit:
            for interact in assembly.fem.interactions.values():
                if interact.name not in initial_step.interactions.keys():
                    initial_step.add_interaction(interact)
                    return

    with open(analysis_dir / "core_input_files/interactions.inp", "w") as d:
        istr = interact_str(assembly.fem)
        if istr != "":
            d.write(istr)
            d.write("\n")



from .elem_formulations import AbaqusDefaultElemTypes


class AbaqusInpFormat:
    def __init__(self):
        self.underline_prefix_is_internal = True


class AbaqusOptions:
    def __init__(self):
        self.default_elements = AbaqusDefaultElemTypes()
        self.inp_format = AbaqusInpFormat()



from __future__ import annotations

from enum import Enum


class BaseEnum(Enum):
    @classmethod
    def from_str(cls, value: str):
        if isinstance(value, cls):
            return value
        key_map = {x.value.lower(): x for x in cls}
        return key_map.get(value.lower(), None)


class GeomRepr(Enum):
    SOLID = "solid"
    SHELL = "shell"
    LINE = "line"

    @staticmethod
    def from_str(value: str) -> GeomRepr:
        keymap = {x.value.lower(): x for x in GeomRepr}
        result = keymap.get(value.lower())
        if result is None:
            raise ValueError(f"Geometric Representation needs to be one of {list(keymap.keys())}")

        return result



from typing import TYPE_CHECKING

from ada.fem import Surface

from .helper_utils import get_instance_name

if TYPE_CHECKING:
    from ada import FEM


def surfaces_str(fem: "FEM", on_assembly_level):
    if len(fem.surfaces) == 0:
        return "** No Surfaces"

    return "\n".join([surface_str(s, on_assembly_level) for s in fem.surfaces.values()])


def surface_str(surface: Surface, write_on_assembly_level: bool) -> str:
    """Surface assignments str"""
    from ada.fem.elements import find_element_type_from_list
    from ada.fem.shapes import ElemType

    top_line = f"*Surface, type={surface.type}, name={surface.name}"

    if surface.id_refs is not None:
        id_refs_str = "\n".join([f"{m[0]}, {m[1]}" for m in surface.id_refs]).strip()
        return f"""{top_line}\n{id_refs_str}"""

    if surface.type == surface.TYPES.NODE:
        elem_face_index_label = surface.weight_factor
    else:
        elem_face_index_label = surface.el_face_index

    fs_str = ""
    if type(surface.fem_set) is not list:
        f_sets = [surface.fem_set]
        el_face_indices = [elem_face_index_label]
    else:
        f_sets = surface.fem_set
        el_face_indices = elem_face_index_label

    for fs, el_f_index in zip(f_sets, el_face_indices):
        set_ref = get_instance_name(fs, write_on_assembly_level)

        if surface.type == surface.TYPES.NODE:
            fs_str += f"{set_ref}\n"
            continue
        el_type = find_element_type_from_list(fs.members)
        if el_type == ElemType.SOLID:
            fs_str += f"{set_ref}, S{el_f_index + 1}\n"
        elif el_type == ElemType.SHELL:
            face_str = "SNEG" if el_f_index == -1 else "SPOS"
            fs_str += f"{set_ref}, {face_str}\n"
        else:
            raise NotImplementedError()

    return f"""{top_line}\n{fs_str.strip()}"""



import os
import shutil
from typing import TYPE_CHECKING

from .templates import main_inp_str
from .write_interactions import interact_str
from .write_predefined_state import predefined_fields_str
from .write_steps import constraint_control, main_step_inp_str

if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly, Part


def write_main_inp_str(assembly: "Assembly", analysis_dir) -> str:
    part_str = "\n".join(map(part_inp_str, filter(skip_if_this, assembly.get_all_subparts())))
    i_str = "\n".join((instance_str(i, analysis_dir) for i in filter(inst_skip, assembly.get_all_subparts()))).rstrip()

    step_str = "** No Steps added"
    incl = "*INCLUDE,INPUT=core_input_files"
    ampl_str = "**"
    consec_str = "**"
    iprop_str = "**"
    int_str = "**"

    if len(assembly.fem.steps) > 0:
        step_str = "\n".join(list(map(main_step_inp_str, assembly.fem.steps))).rstrip()
    if len(assembly.fem.amplitudes) > 0:
        ampl_str = f"{incl}\\amplitude_data.inp"
    if len(assembly.fem.connector_sections) > 0:
        consec_str = f"{incl}\\connector_sections.inp"
    if len(assembly.fem.intprops) > 0:
        iprop_str = f"{incl}\\interaction_prop.inp"
    if interact_str(assembly.fem) != "" or predefined_fields_str(assembly.fem) != "":
        int_str = f"{incl}\\interactions.inp"

    mat_str = f"{incl}\\materials.inp"
    fix_str = f"{incl}\\bc_data.inp"

    return main_inp_str.format(
        part_str=part_str,
        instance_str=i_str,
        mat_str=mat_str,
        fix_str=fix_str,
        step_str=step_str,
        ampl_str=ampl_str,
        consec_str=consec_str,
        int_prop_str=iprop_str,
        interact_str=int_str,
        constr_ctrl=constraint_control(assembly.fem),
    )


def part_inp_str(part: "Part") -> str:
    return """**\n*Part, name={name}\n*INCLUDE,INPUT=bulk_{name}\\{inp_file}\n*End Part\n**""".format(
        name=part.name, inp_file="aba_bulk.inp"
    )


def instance_str(part: "Part", analysis_dir) -> str:
    if part.fem.initial_state is None:
        return f"""**\n*Instance, name={part.fem.instance_name}, part={part.name}\n*End Instance"""

    istep = part.fem.initial_state
    analysis_name = os.path.basename(istep.initial_state_file.replace(".inp", ""))
    source_dir = os.path.dirname(istep.initial_state_file)
    for f in os.listdir(source_dir):
        if analysis_name in f:
            dest_file = os.path.join(analysis_dir, os.path.basename(f))
            shutil.copy(os.path.join(source_dir, f), dest_file)
    return f"""*Instance, library={analysis_name}, instance={istep.initial_state_part.fem.instance_name}
**
** PREDEFINED FIELD
**
** Name: {part.fem.initial_state.name}   Type: Initial State
*Import, state=yes, update=no
*End Instance"""


def skip_if_this(p):
    if p.fem.initial_state is not None:
        return False
    return len(p.fem.elements)


def inst_skip(p):
    if p.fem.initial_state is not None:
        return True
    return len(p.fem.elements)



from itertools import chain

from ada.concepts.containers import Materials
from ada.core.utils import roundoff
from ada.fem.formats.utils import str_to_int
from ada.materials import Material
from ada.materials.metals import CarbonSteel

from . import cards


def get_materials(bulk_str, part) -> Materials:
    """
    Interpret Material bulk string to FEM objects


    TDMATER: Material Element
    MISOSEL: linear elastic,isotropic

    TDMATER   4.00000000E+00  4.50000000E+01  1.07000000E+02  0.00000000E+00
            softMat

    MISOSEL   1.00000000E+00  2.10000003E+11  3.00000012E-01  1.15515586E+04
              1.14999998E+00  1.20000004E-05  1.00000000E+00  3.55000000E+08

    :return:
    """

    mat_names = {matid: mat_name for matid, mat_name in map(grab_name, cards.TDMATER.to_ff_re().finditer(bulk_str))}

    misosel = (get_mat(m, mat_names, part) for m in cards.MISOSEL.to_ff_re().finditer(bulk_str))
    morsmel = (get_morsmel(m, mat_names, part) for m in cards.MORSMEL.to_ff_re().finditer(bulk_str))
    return Materials(chain(misosel, morsmel), parent=part)


def grab_name(m):
    d = m.groupdict()
    return str_to_int(d["geo_no"]), d["name"]


def get_morsmel(m, mat_names, part) -> Material:
    """
    MORSMEL

    Anisotropy, Linear Elastic Structural Analysis, 2-D Membrane Elements and 2-D Thin Shell Elements

    :param m:
    :return:
    """

    d = m.groupdict()
    matno = str_to_int(d["matno"])

    mat_model = CarbonSteel(
        rho=roundoff(d["rho"]),
        E=roundoff(d["d11"]),
        v=roundoff(d["ps1"]),
        alpha=roundoff(d["alpha1"]),
        zeta=roundoff(d["damp1"]),
        sig_y=5e6,
    )

    return Material(name=mat_names[matno], mat_id=matno, mat_model=mat_model, metadata=d, parent=part)


def get_mat(match, mat_names, part) -> Material:
    d = match.groupdict()
    matno = str_to_int(d["matno"])
    mat_model = CarbonSteel(
        rho=roundoff(d["rho"]),
        E=roundoff(d["young"]),
        v=roundoff(d["poiss"]),
        alpha=roundoff(d["damp"]),
        zeta=roundoff(d["alpha"]),
        sig_y=roundoff(d["yield"]),
    )
    return Material(name=mat_names[matno], mat_id=matno, mat_model=mat_model, parent=part)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import Assembly, Material


def materials_str(assembly: "Assembly"):
    all_mat = [mat for p in assembly.get_all_parts_in_assembly(True) for mat in p.materials]
    return "\n".join([material_str(mat) for mat in all_mat])


def material_str(material: "Material") -> str:
    if "aba_inp" in material.metadata.keys():
        return material.metadata["aba_inp"]

    no_compression = material.metadata["no_compression"] if "no_compression" in material.metadata.keys() else False
    compr_str = "\n*No Compression" if no_compression is True else ""

    pl_str = ""
    if material.model.plasticity_model is not None:
        pl_model = material.model.plasticity_model
        if pl_model.eps_p is not None and len(pl_model.eps_p) != 0:
            pl_str = "\n*Plastic\n"
            pl_str += "\n".join(
                ["{x:>12.5E}, {y:>10}".format(x=x, y=y) for x, y in zip(pl_model.sig_p, pl_model.eps_p)]
            )

    alpha = material.model.rayleigh_damping.alpha
    beta = material.model.rayleigh_damping.beta
    d_str = ""
    if alpha is not None and beta is not None:
        d_str = f"\n*Damping, alpha={alpha}, beta={beta}"

    exp_str = ""
    if material.model.zeta is not None and material.model.zeta != 0.0:
        exp_str = f"\n*Expansion\n {material.model.zeta}"

    # Density == 0.0 is unsupported
    density = material.model.rho if material.model.rho > 0.0 else 1e-6

    return f"""*Material, name={material.name}
*Elastic
{material.model.E:.6E},  {material.model.v}{compr_str}
*Density
{density},{exp_str}{d_str}{pl_str}"""



from __future__ import annotations

import os
import pathlib
import pickle
import shutil
import subprocess
from typing import TYPE_CHECKING

import numpy as np

from ada.config import get_logger

if TYPE_CHECKING:
    from ada.fem.results.common import FEAResult, FieldData, Mesh

_script_dir = pathlib.Path(__file__).parent.resolve().absolute()


ABA_IO = _script_dir / "aba_io.py"
logger = get_logger()


def convert_to_pckle(odb_path, pickle_path, use_aba_version=None):
    aba_ver = "abaqus" if use_aba_version is None else use_aba_version
    aba_exe_path = pathlib.Path(shutil.which(aba_ver))

    odb_path = pathlib.Path(odb_path)

    if os.path.isfile(pickle_path):
        os.remove(pickle_path)

    logger.info(f'Extracting ODB data from "{odb_path.name}" using Abaqus/Python')

    backup_odb = odb_path.parent / f"{odb_path.stem}_backup.odb"
    if backup_odb.exists() is False:
        logger.info(f'Copying a backup of the odb file to "{backup_odb}" in case python corrupts the odb file')
        shutil.copy(odb_path, backup_odb)

    res = subprocess.run([aba_exe_path, "python", ABA_IO, odb_path], cwd=ABA_IO.parent, capture_output=True)
    logger.info(str(res.stdout, encoding="utf-8"))
    stderr = str(res.stderr, encoding="utf-8")
    if stderr != "":
        logger.error(stderr)


def get_odb_data(odb_path, overwrite=False, use_aba_version=None):
    odb_path = pathlib.Path(odb_path)
    pickle_path = odb_path.with_suffix(".pckle")

    if pickle_path.exists() is False or overwrite is True:
        convert_to_pckle(odb_path, pickle_path, use_aba_version)

    with open(pickle_path, "rb") as f:
        data = pickle.load(f)

    return data


def read_odb_pckle_file(pickle_path: str | pathlib.Path) -> FEAResult:
    from ada.fem.formats.general import FEATypes
    from ada.fem.results.common import FEAResult

    if isinstance(pickle_path, pathlib.Path) is False:
        pickle_path = pathlib.Path(pickle_path)

    if pickle_path.suffix.lower() == ".odb":
        pickle_path = pickle_path.with_suffix(".pckle")

    with open(pickle_path, "rb") as f:
        data = pickle.load(f)

    mesh = get_odb_instance_data(data["rootAssembly"]["instances"])
    fields = get_odb_frame_data(data["steps"])

    return FEAResult(
        name=pickle_path.stem, software=FEATypes.ABAQUS, mesh=mesh, results=fields, results_file_path=pickle_path
    )


def get_odb_field_data(field_name, field_data, frame_num):
    from ada.fem.results.field_data import ElementFieldData, NodalFieldData

    field_type, components, data = field_data

    if field_type == "ELEMENT_NODAL":
        field_values = np.array(list(yield_elem_nodal_data(data)))
        return ElementFieldData(field_name, frame_num, components, values=field_values)
    elif field_type == "NODAL":
        field_values = np.array(list(yield_nodal_data(data)))
        return NodalFieldData(field_name, frame_num, components, field_values)
    else:
        raise NotImplementedError()


def get_odb_frame_data(steps: dict) -> list[FieldData]:
    frame_num = 0
    fields = []
    for step_name, step_data in dict(sorted(steps.items(), key=lambda x: x[1]["totalTime"])).items():
        for frame in step_data["frames"]:
            for key, value in frame.items():
                field = get_odb_field_data(key, value["values"], frame_num)
                fields.append(field)
            frame_num += 1

    return fields


def get_odb_instance_data(instances) -> Mesh:
    from ada.fem.formats.abaqus.elem_shapes import abaqus_el_type_to_ada
    from ada.fem.formats.general import FEATypes
    from ada.fem.results.common import ElementBlock, ElementInfo, FemNodes, Mesh

    if len(instances) > 1:
        raise NotImplementedError("Multi-instances results are not yet supported")

    instance = instances[0]

    ids, coords = zip(*instance["nodes"])
    el_ids, el_type_array, nodes_connectivity, sec_cat = zip(*instance["elements"])
    el_type_set = set(el_type_array)
    if len(el_type_set) != 1:
        raise NotImplementedError("Mixed element sets not yet supported")

    el_type = el_type_array[0]
    shape = abaqus_el_type_to_ada(el_type)
    elem_info = ElementInfo(type=shape, source_software=FEATypes.ABAQUS, source_type=el_type)
    el_block = ElementBlock(
        elem_info=elem_info, node_refs=np.array(nodes_connectivity, dtype=int), identifiers=np.array(el_ids, dtype=int)
    )
    el_blocks = [el_block]
    nodes = FemNodes(coords=np.array(coords, dtype=float), identifiers=np.array(ids, dtype=int))

    return Mesh(elements=el_blocks, nodes=nodes)


def yield_elem_nodal_data(data):
    for x in data:
        spn = x["sec_p_num"]
        if isinstance(spn, dict):
            spn = -1
        if isinstance(x["data"], list) is False:
            x["data"] = [x["data"]]

        yield x["elementLabel"], spn, x["nodeLabel"], *x["data"]


def yield_nodal_data(data):
    for x in data:
        yield x[0], *x[1]



from __future__ import annotations

import pathlib

from ada.fem.formats.abaqus.results.read_odb import (
    convert_to_pckle,
    read_odb_pckle_file,
)
from ada.fem.formats.sesam.results.read_sif import read_sif_file
from ada.fem.formats.sesam.results.sin2sif import convert_sin_to_sif
from ada.fem.results.common import FEAResult


def from_results_file(fem_res: str | pathlib.Path, fem_format: str = None, force_conversion=False) -> FEAResult:
    file_ref = pathlib.Path(fem_res)
    suffix = file_ref.suffix.lower()

    if suffix == ".sin":
        sif_file = file_ref.with_suffix(".sif")
        if sif_file.exists() and force_conversion is False:
            return read_sif_file(sif_file)

        convert_sin_to_sif(file_ref)
        return read_sif_file(sif_file)
    elif suffix == ".sif":
        return read_sif_file(file_ref.with_suffix(".sif"))
    elif suffix == ".odb":
        pckl_data = file_ref.with_suffix(".pckle")
        if pckl_data.exists() is False:
            convert_to_pckle(file_ref, pckl_data)
        return read_odb_pckle_file(pckl_data)
    else:
        raise NotImplementedError()



from typing import TYPE_CHECKING

import numpy as np

from ada.config import get_logger
from ada.fem.shapes import definitions as shape_def

from ..common import ada_to_med_type
from .write_sets import _add_cell_sets

if TYPE_CHECKING:
    from ada.concepts.spatial import Part

logger = get_logger()


def elements_str(part: "Part", time_step, profile, families):
    """

    Add the following ['FAM', 'NOD', 'NUM'] to the 'MAI' group

    **NOD** requires 'CGT' and 'NBR' attrs

    :param part:
    :param time_step:
    :param profile:
    :param families:
    :return:
    """

    def get_node_ids_from_element(el_):
        return [int(n.id) for n in el_.nodes]

    elements_group = time_step.create_group("MAI")
    elements_group.attrs.create("CGT", 1)
    for group, elements in part.fem.elements.group_by_type():
        if isinstance(group, (shape_def.MassTypes, shape_def.SpringTypes)):
            logger.warning("NotImplemented: Skipping Mass or Spring Elements")
            continue
        med_type = ada_to_med_type(group)
        elements = list(elements)
        cells = np.array(list(map(get_node_ids_from_element, elements)))

        med_cells = elements_group.create_group(med_type)
        med_cells.attrs.create("CGT", 1)
        med_cells.attrs.create("CGS", 1)
        med_cells.attrs.create("PFL", np.string_(profile))

        nod = med_cells.create_dataset("NOD", data=cells.flatten(order="F"))
        nod.attrs.create("CGT", 1)
        nod.attrs.create("NBR", len(cells))

        # Node Numbering is necessary for proper handling of
        num = med_cells.create_dataset("NUM", data=[int(el.id) for el in elements])
        num.attrs.create("CGT", 1)
        num.attrs.create("NBR", len(cells))

    # Add Element sets
    if len(part.fem.elsets.keys()) > 0:
        _add_cell_sets(elements_group, part, families)



from .read_beams import get_beams
from .read_joints import get_joints
from .read_materials import get_materials
from .read_sections import get_sections

__all__ = ["get_materials", "get_sections", "get_joints", "get_beams"]



from typing import TYPE_CHECKING

import numpy as np

from ada.config import get_logger
from ada.fem.shapes import definitions as shape_def

from ..common import ada_to_med_type
from .helper_utils import resolve_ids_in_multiple

if TYPE_CHECKING:
    from ada.concepts.spatial import Part

logger = get_logger()


def _add_cell_sets(cells_group, part: "Part", families):
    """

    :param cells_group:
    :param part:
    :param families:
    """
    cell_id_num = -4

    element = families.create_group("ELEME")
    tags = dict()
    tags_data = dict()

    cell_id_current = cell_id_num
    for cell_set in part.fem.elsets.values():
        tags[cell_id_current] = [cell_set.name]
        tags_data[cell_id_current] = cell_set.members
        cell_id_current -= 1

    res_data = resolve_ids_in_multiple(tags, tags_data, True)

    def get_node_ids_from_element(el_):
        return [int(n.id - 1) for n in el_.nodes]

    for group, elements in part.fem.elements.group_by_type():
        if isinstance(group, (shape_def.MassTypes, shape_def.SpringTypes)):
            logger.warning("NotImplemented: Skipping Mass or Spring Elements")
            continue
        elements = list(elements)
        cell_ids = {el.id: i for i, el in enumerate(elements)}

        cell_data = np.zeros(len(elements), dtype=np.int32)

        for t, mem in res_data.items():
            list_filtered = [cell_ids[el.id] for el in filter(lambda x: x.type == group, mem)]
            for index in list_filtered:
                cell_data[index] = t

        cells = np.array(list(map(get_node_ids_from_element, elements)))
        med_type = ada_to_med_type(group)
        med_cells = cells_group.get(med_type)
        family = med_cells.create_dataset("FAM", data=cell_data)
        family.attrs.create("CGT", 1)
        family.attrs.create("NBR", len(cells))

    _write_families(element, tags)


def _add_node_sets(nodes_group, part: "Part", points, families):
    """
    :param nodes_group:
    :param part:
    :param families:
    """
    tags = dict()
    nsets = dict()
    for key, val in part.fem.nsets.items():
        nsets[key] = [int(p.id) for p in val]

    points = _set_to_tags(nsets, points, 2, tags)

    family = nodes_group.create_dataset("FAM", data=points)
    family.attrs.create("CGT", 1)
    family.attrs.create("NBR", len(points))

    # For point tags
    node = families.create_group("NOEUD")
    _write_families(node, tags)


def _resolve_element_in_use_by_other_set(tagged_data, ind, tags, name, is_elem):
    existing_id = int(tagged_data[ind])
    current_tags = tags[existing_id]
    all_tags = current_tags + [name]

    if name in current_tags:
        logger.error("Unexpected error. Name already exists in set during resolving set members.")

    new_int = None
    for i_, t_ in tags.items():
        if all_tags == t_:
            new_int = i_
            break

    if new_int is None:
        new_int = int(min(tags.keys()) - 1) if is_elem else int(max(tags.keys()) + 1)
        tags[new_int] = tags[existing_id] + [name]

    tagged_data[ind] = new_int


def _set_to_tags(sets, data, tag_start_int, tags, id_map=None):
    """

    :param sets:
    :param data:
    :param tag_start_int:
    :param
    :return: The tagged data.
    """
    tagged_data = np.zeros(len(data), dtype=np.int32)
    tag_int = 0 + tag_start_int

    is_elem = False if tag_int > 0 else True

    tag_int = tag_start_int
    tag_map = dict()
    # Generate basic tags upfront
    for name in sets.keys():
        tags[tag_int] = [name]
        tag_map[name] = tag_int
        if is_elem is True:
            tag_int -= 1
        else:
            tag_int += 1

    for name, set_data in sets.items():
        if len(set_data) == 0:
            continue

        for index_ in set_data:
            index = int(index_ - 1)

            if id_map is not None:
                index = id_map[index_]

            if index > len(tagged_data) - 1:
                raise IndexError()

            if tagged_data[index] != 0:  # id is already defined in another set
                _resolve_element_in_use_by_other_set(tagged_data, index, tags, name, is_elem)
            else:
                tagged_data[index] = tag_map[name]

    return tagged_data


def _family_name(set_id, name):
    """Return the FAM object name corresponding to the unique set id and a list of subset names"""
    return "FAM" + "_" + str(set_id) + "_" + "_".join(name)


def _write_families(fm_group, tags):
    """Write point/cell tag information under FAS/[mesh_name]"""
    for set_id, name in tags.items():
        family = fm_group.create_group(_family_name(set_id, name))
        family.attrs.create("NUM", set_id)
        group = family.create_group("GRO")
        group.attrs.create("NBR", len(name))  # number of subsets
        dataset = group.create_dataset("NOM", (len(name),), dtype="80int8")
        for i in range(len(name)):
            # make name 80 characters
            name_80 = name[i] + "\x00" * (80 - len(name[i]))
            # Needs numpy array, see <https://github.com/h5py/h5py/issues/1735>
            dataset[i] = np.array([ord(x) for x in name_80])



from __future__ import annotations

import os
import pathlib
import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

from .write_bcs import add_boundary_conditions
from .write_beams import add_beams
from .write_masses import add_masses
from .write_materials import add_materials
from .write_sat_embedded import embed_sat_geometry
from .write_sections import add_sections

if TYPE_CHECKING:
    from ada import Part

_XML_TEMPLATE = pathlib.Path(__file__).parent / "resources/xml_blank.xml"


def write_xml(part: Part, xml_file, embed_sat=False):
    if not isinstance(xml_file, pathlib.Path):
        xml_file = pathlib.Path(xml_file)

    tree = ET.parse(_XML_TEMPLATE)
    root = tree.getroot()

    part.consolidate_sections()
    part.consolidate_materials()

    # Find the <properties> element
    structure_domain = root.find("./model/structure_domain")
    structures_elem = ET.SubElement(structure_domain, "structures")
    properties = structure_domain.find("./properties")

    # Add Properties
    add_sections(properties, part)
    add_materials(properties, part)

    # Add SAT geometry (maybe only applicable for plate geometry)
    sat_map = dict()
    if embed_sat:
        sat_map = embed_sat_geometry(structure_domain, part)

    # Add structural elements
    add_beams(structures_elem, part, sat_map)
    add_boundary_conditions(structures_elem, part)
    add_masses(structures_elem, part)

    # Write the modified XML back to the file
    os.makedirs(xml_file.parent, exist_ok=True)
    tree.write(str(xml_file))



import re
from typing import TYPE_CHECKING

from ada.core.utils import Counter
from ada.fem import Mass
from ada.fem.containers import FemElements
from ada.fem.formats.utils import str_to_int
from ada.fem.shapes import definitions as shape_def

from .helper_utils import _re_in, get_set_from_assembly

if TYPE_CHECKING:
    from ada import FEM


def get_mass_from_bulk(bulk_str, parent: "FEM") -> FemElements:
    """

    *MASS,ELSET=MASS3001
    2.00000000E+03,

    :return:
    """
    mass_ids = Counter(int(parent.elements.max_el_id + 1))

    re_masses = re.compile(
        r"\*(?P<mass_type>Nonstructural Mass|Mass|Rotary Inertia),\s*elset=(?P<elset>.*?)"
        r"(?:,\s*type=(?P<ptype>.*?)\s*|\s*)(?:, units=(?P<units>.*?)|\s*)\n\s*(?P<mass>.*?)$",
        _re_in,
    )
    return FemElements((get_mass(m, parent, mass_ids) for m in re_masses.finditer(bulk_str)), fem_obj=parent)


aba_to_ada_mass_map = {"ROTARY INERTIA": shape_def.MassTypes.ROTARYI, "MASS": shape_def.MassTypes.MASS}
ada_to_aba_mass_map = {val: key for key, val in aba_to_ada_mass_map.items()}


def get_mass(match, parent: "FEM", mass_id_gen):
    d = match.groupdict()
    elset = get_set_from_assembly(d["elset"], parent, "elset")
    mass_type = d["mass_type"]
    mass_type_general = aba_to_ada_mass_map.get(mass_type.upper(), None)
    if mass_type_general is None:
        raise NotImplementedError(f'Mass type "{mass_type}" is not yet supported by general ADA')

    p_type = d["ptype"]
    mass_ints = [str_to_int(x.strip()) for x in d["mass"].split(",") if x.strip() != ""]
    if len(mass_ints) == 1:
        mass_ints = mass_ints[0]
    units = d["units"]
    elem = elset.members[0]
    mass = Mass(
        d["elset"], elset, mass_ints, mass_type_general, p_type, mass_id=next(mass_id_gen), units=units, parent=parent
    )
    elem.mass_prop = mass
    return mass



from __future__ import annotations

import os
import pathlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable, Iterable

import ifcopenshell
import ifcopenshell.geom

from ada.base.changes import ChangeAction
from ada.config import get_logger
from ada.ifc.utils import assembly_to_ifc_file, default_settings, get_unit_type
from ada.ifc.write.write_sections import get_profile_class
from ada.ifc.write.write_user import create_owner_history_from_user

if TYPE_CHECKING:
    from OCC.Core.TopoDS import TopoDS_Compound, TopoDS_Shape

    from ada import Assembly, Section, User
    from ada.ifc.read.read_ifc import IfcReader
    from ada.ifc.write.write_ifc import IfcWriter

logger = get_logger()


@dataclass
class IfcStore:
    ifc_file_path: pathlib.Path | os.PathLike = None
    assembly: Assembly = None
    settings: ifcopenshell.geom.settings = field(default_factory=default_settings)

    f: ifcopenshell.file = None
    owner_history: ifcopenshell.entity_instance = None
    writer: IfcWriter = None
    reader: IfcReader = None
    callback: Callable[[int, int], None] | None = None

    def __post_init__(self):
        if self.f is None:
            if self.ifc_file_path is not None:
                self.ifc_file_path = pathlib.Path(self.ifc_file_path)
                if self.ifc_file_path.exists():
                    self.f = ifcopenshell.open(str(self.ifc_file_path))
            elif self.assembly is not None:
                self.f = assembly_to_ifc_file(self.assembly)
                self.add_standard_contexts()

    def add_standard_contexts(self):
        contexts = list(self.f.by_type("IfcGeometricRepresentationContext"))
        model_context = list(filter(lambda x: x.ContextType.upper() == "MODEL", contexts))[0]
        for cid, target_view in [("Body", "MODEL_VIEW"), ("Axis", "GRAPH_VIEW"), ("Box", "MODEL_VIEW")]:
            self.f.create_entity(
                "IfcGeometricRepresentationSubContext",
                ContextIdentifier=cid,
                ContextType="Model",
                ParentContext=model_context,
                TargetView=target_view,
            )
        plan_context = self.f.create_entity(
            "IfcGeometricRepresentationContext",
            ContextType="Plan",
            CoordinateSpaceDimension=2,
            Precision=1e-5,
            WorldCoordinateSystem=model_context.WorldCoordinateSystem,
        )
        for cid, target_view in [
            ("Axis", "GRAPH_VIEW"),
            ("Annotation", "PLAN_VIEW"),
            ("Annotation", "SECTION_VIEW"),
            ("Annotation", "ELEVATION_VIEW"),
        ]:
            self.f.create_entity(
                "IfcGeometricRepresentationSubContext",
                ContextIdentifier=cid,
                ContextType="Plan",
                ParentContext=plan_context,
                TargetView=target_view,
            )

    def update_owner(self, user: User):
        self.owner_history = create_owner_history_from_user(user, self.f)

    def get_context(self, context_id):
        contexts = list(self.f.by_type("IfcGeometricRepresentationContext"))
        subcontexts = list(self.f.by_type("IfcGeometricRepresentationSubContext"))
        if len(contexts) == 1 and len(subcontexts) == 0:
            return contexts[0]

        contexts = [x for x in contexts if x.ContextIdentifier == context_id and x.ContextType == "Model"]
        if len(contexts) == 0:
            raise ValueError(f'0 IfcGeometry Subcontexts found with "{context_id=}"')
        if len(contexts) > 1:
            raise ValueError(f'Multiple Subcontexts found with "{context_id=}"')

        return contexts[0]

    def sync(self, include_fem=False, progress_callback: Callable[[int, int], None] = None):
        from ada.ifc.write.write_ifc import IfcWriter

        self.writer = IfcWriter(self)
        self.writer.callback = progress_callback
        a = self.assembly

        a.consolidate_sections()
        a.consolidate_materials()

        self.update_owner(a.user)

        num_new_spatial_objects = self.writer.sync_spatial_hierarchy(include_fem=include_fem)

        self.writer.sync_sections()
        self.writer.sync_materials()

        num_new_objects = self.writer.sync_added_physical_objects()

        self.writer.sync_added_welds()

        self.writer.sync_mapped_instances()

        num_mod = self.writer.sync_modified_physical_objects()

        self.writer.sync_presentation_layers()

        num_del = self.writer.sync_deleted_physical_objects()

        add_str = f"Added {num_new_objects} objects and {num_new_spatial_objects} spatial elements"
        mod_str = f"Modified {num_mod} objects"
        del_str = f"Deleted {num_del} objects"

        print(f"Sync Complete. {add_str}. {mod_str}. {del_str}")
        self.callback = None

    def save_to_file(self, filepath: str | os.PathLike):
        with open(filepath, "w") as f:
            f.write(self.f.wrapped_data.to_string())

    def load_ifc_content_from_file(
        self, ifc_file: str | os.PathLike | ifcopenshell.file = None, data_only=False, elements2part=None
    ) -> None:
        from ada.ifc.read.read_ifc import IfcReader

        if self.ifc_file_path is None:
            if ifc_file is None:
                raise ValueError("No ifc file is attached")
            if isinstance(ifc_file, (str, os.PathLike)):
                self.ifc_file_path = ifc_file
                self.f = IfcStore.ifc_obj_from_ifc_file(ifc_file)
            else:
                self.f = ifc_file

        if self.assembly is None:
            raise ValueError("Assembly must be attached before loading IFC content")

        self.reader = IfcReader(self)

        target_units = None
        unit_type = get_unit_type(self.f)

        if unit_type != self.assembly.units:
            target_units = self.assembly.units
            self.assembly.units = unit_type

        if elements2part is None:
            self.reader.load_spatial_hierarchy()

        # Load Materials
        self.reader.load_materials()

        # Load physical elements
        self.reader.load_objects(data_only=data_only, elements2part=elements2part)

        if target_units is not None:
            self.assembly.units = target_units

        self.reader.load_presentation_layers()

        ifc_file_name = "object" if self.ifc_file_path is None else self.ifc_file_path

        for obj in self.assembly.get_all_sections():
            obj.change_type = ChangeAction.NOCHANGE

        for obj in self.assembly.get_all_materials():
            obj.change_type = ChangeAction.NOCHANGE

        for obj in self.assembly.get_all_physical_objects():
            obj.change_type = ChangeAction.NOCHANGE

        for obj in self.assembly.get_all_parts_in_assembly(include_self=True):
            obj.change_type = ChangeAction.NOCHANGE

        print(f'Import of IFC file "{ifc_file_name}" is complete')

    def get_ifc_geom(self, ifc_elem, settings: ifcopenshell.geom.settings):
        return ifcopenshell.geom.create_shape(settings, inst=ifc_elem)

    def get_ifc_geom_iterator(self, settings: ifcopenshell.geom.settings, cpus: int = None):
        import multiprocessing

        products = None
        if self.assembly is not None:
            products = []
            for x in self.assembly.get_all_physical_objects(pipe_to_segments=True):
                try:
                    product = self.f.by_guid(x.guid)
                except RuntimeError as e:
                    raise RuntimeError(e)
                products.append(product)
        cpus = multiprocessing.cpu_count() if cpus is None else cpus
        return ifcopenshell.geom.iterator(settings, self.f, cpus, include=products)

    def iter_occ_shapes_from_ifc(
        self, settings: ifcopenshell.geom.settings = None
    ) -> Iterable[tuple[str, TopoDS_Shape | TopoDS_Compound]]:
        if settings is None:
            settings = ifcopenshell.geom.settings()
            settings.set(settings.USE_PYTHON_OPENCASCADE, True)
        iterator = self.get_ifc_geom_iterator(settings)
        iterator.initialize()
        while True:
            shape = iterator.get()
            if shape:
                if hasattr(shape, "geometry"):
                    name = shape[0].name
                    yield name, shape.geometry
                else:
                    logger.warning(f"Shape {shape} is not a TopoDS_Shape")

            if not iterator.next():
                break

    def get_by_guid(self, guid: str) -> ifcopenshell.entity_instance:
        return self.f.by_guid(guid)

    def get_beam_type(self, section: Section, match_description=False) -> ifcopenshell.entity_instance:
        for beam_type in self.f.by_type("IfcBeamType"):
            if section.name == beam_type.Name:
                return beam_type

    def get_profile_def(self, section: Section) -> ifcopenshell.entity_instance:
        profile_class = get_profile_class(section)
        for profile_def in self.f.by_type(profile_class.get_ifc_type()):
            if profile_def.ProfileName == section.name:
                return profile_def

    @staticmethod
    def from_ifc(ifc_file: str | os.PathLike | ifcopenshell.file, make_a_copy=True) -> IfcStore:
        ifc_file_path = None

        if isinstance(ifc_file, (str, os.PathLike)):
            ifc_file_path = ifc_file
            f = IfcStore.ifc_obj_from_ifc_file(ifc_file)
        else:
            if make_a_copy:
                f = IfcStore.copy_ifc_obj(ifc_file)
            else:
                f = ifc_file

        return IfcStore(ifc_file_path=ifc_file_path, f=f)

    @staticmethod
    def ifc_obj_from_ifc_file(ifc_file: str | os.PathLike) -> ifcopenshell.file:
        ifc_file = pathlib.Path(ifc_file).resolve().absolute()
        if ifc_file.exists() is False:
            raise FileNotFoundError(f'Unable to find "{ifc_file}"')
        return ifcopenshell.open(str(ifc_file))

    @staticmethod
    def copy_ifc_obj(ifc_file: ifcopenshell.file) -> ifcopenshell.file:
        return ifcopenshell.file.from_string(ifc_file.wrapped_data.to_string())



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING

from ada.config import get_logger
from ada.fem import StepEigen
from ada.fem.formats.utils import DatFormatReader

from .sin2sif import convert_sin_to_sif

if TYPE_CHECKING:
    from ada.fem.results import Results
    from ada.fem.results.eigenvalue import EigenDataSummary

logger = get_logger()


def get_eigen_data(dat_file: str | os.PathLike) -> EigenDataSummary:
    from ada.fem.results.eigenvalue import EigenDataSummary, EigenMode

    dtr = DatFormatReader()

    re_compiled = dtr.compile_ff_re([int] + [float] * 3, separator=";")

    eig_str = "printofeigenvalues"
    eig_res = dtr.read_data_lines(dat_file, re_compiled, eig_str, split_data=True)
    eigen_modes: list[EigenMode] = []

    # Note! participation factors and effective modal mass are each deconstructed into 6 degrees of freedom
    for mode, eig_value, eig_freq, period in eig_res:
        eig_output = dict(
            eigenvalue=eig_value.replace(";", ""),
            f_hz=eig_freq.replace(";", ""),
        )
        eigen_modes.append(EigenMode(no=mode.replace(";", ""), **eig_output))

    return EigenDataSummary(eigen_modes)


def read_sesam_results(results: "Results", file_ref: pathlib.Path, overwrite):
    dat_file = (file_ref.parent / "SESTRA").with_suffix(".LIS")
    if dat_file.exists() and type(results.assembly.fem.steps[0]) == StepEigen:
        results.eigen_mode_data = get_eigen_data(dat_file)

    convert_sin_to_sif(results.results_file_path)

    logger.error("Result mesh data extraction is not supported for sesam")
    return None



from __future__ import annotations

from enum import Enum


class SectionTypes(Enum):
    IfcIShapeProfileDef = "IfcIShapeProfileDef"
    IfcArbitraryClosedProfileDef = "IfcArbitraryClosedProfileDef"
    IfcArbitraryProfileDefWithVoids = "IfcArbitraryProfileDefWithVoids"
    IfcCircleProfileDef = "IfcCircleProfileDef"
    IfcCircleHollowProfileDef = "IfcCircleHollowProfileDef"
    IfcUShapeProfileDef = "IfcUShapeProfileDef"

    @staticmethod
    def from_str(class_name: str):
        key_map = {x.value.lower(): x for x in SectionTypes}
        return key_map.get(class_name.lower())


class SpatialTypes(Enum):
    # Spatial Top Element
    IfcSite = "IfcSite"
    # Spatial SubElements
    IfcBuilding = "IfcBuilding"
    IfcSpatialZone = "IfcSpatialZone"
    IfcBuildingStorey = "IfcBuildingStorey"
    IfcElementAssembly = "IfcElementAssembly"
    IfcGrid = "IfcGrid"

    @staticmethod
    def from_str(class_name: str):
        key_map = {x.value.lower(): x for x in SpatialTypes}
        return key_map.get(class_name.lower())

    @staticmethod
    def is_valid_spatial_type(ifc_class: str | SpatialTypes) -> bool:
        if isinstance(ifc_class, str):
            ifc_class = SpatialTypes.from_str(ifc_class)

        return ifc_class in list(SpatialTypes)


class ShapeTypes(Enum):
    IfcBuildingElementProxy = "IfcBuildingElementProxy"
